/* check.c */

#include <Pilot.h>

#include "life.h"


/*
 * constants
 */

#define GLIDER_LENGTH 5
#define GLIDER_PERIOD 4

#define GLIDER_MARGIN 6

struct glider {
    cell_t adjust[GLIDER_PERIOD][GLIDER_LENGTH];
    cell_t next;
};

static struct glider glider_up_left = {
  {
    { ADJUST(0, 0), ADJUST(1, 0), ADJUST(2, 0), ADJUST(0, 1), ADJUST(1, 2) },
    { ADJUST(1, -1), ADJUST(0, 0), ADJUST(1, 0), ADJUST(0, 1), ADJUST(2, 1) },
    { ADJUST(0, -1), ADJUST(1, -1), ADJUST(0, 0), ADJUST(2, 0), ADJUST(0, 1) },
    { ADJUST(0, -1), ADJUST(1, -1), ADJUST(-1, 0), ADJUST(0, 0), ADJUST(1, 1) }
  },
    ADJUST(-1, -1)
};

static struct glider glider_up_right = {
  {
    { ADJUST(0, 0), ADJUST(1, 0), ADJUST(2, 0), ADJUST(2, 1), ADJUST(1, 2) },
    { ADJUST(1, -1), ADJUST(1, 0), ADJUST(2, 0), ADJUST(0, 1), ADJUST(2, 1) },
    { ADJUST(1, -1), ADJUST(2, -1), ADJUST(0, 0), ADJUST(2, 0), ADJUST(2, 1) },
    { ADJUST(1, -1), ADJUST(2, -1), ADJUST(2, 0), ADJUST(3, 0), ADJUST(1, 1) }
  },
    ADJUST(1, -1)
};

static struct glider glider_down_left = {
  {
    { ADJUST(2, 0), ADJUST(1, 0), ADJUST(0, 0), ADJUST(0, -1), ADJUST(1, -2) },
    { ADJUST(1, 1), ADJUST(1, 0), ADJUST(0, 0), ADJUST(2, -1), ADJUST(0, -1) },
    { ADJUST(1, 1), ADJUST(0, 1), ADJUST(2, 0), ADJUST(0, 0), ADJUST(0, -1) },
    { ADJUST(1, 1), ADJUST(0, 1), ADJUST(0, 0), ADJUST(-1, 0), ADJUST(1, -1) }
  },
    ADJUST(-1, 1)
};

static struct glider glider_down_right = {
  {
    { ADJUST(2, 0), ADJUST(1, 0), ADJUST(0, 0), ADJUST(2, -1), ADJUST(1, -2) },
    { ADJUST(1, 1), ADJUST(2, 0), ADJUST(1, 0), ADJUST(2, -1), ADJUST(0, -1) },
    { ADJUST(2, 1), ADJUST(1, 1), ADJUST(2, 0), ADJUST(0, 0), ADJUST(2, -1) },
    { ADJUST(2, 1), ADJUST(1, 1), ADJUST(3, 0), ADJUST(2, 0), ADJUST(1, -1) }
  },
    ADJUST(1, 1)
};


/*
 * functions
 */

static int is_glider(
    const cell_t *ptr, Boolean up, const struct glider *glider,
    cell_t *origin_ptr)
{
    int i, j;

    for (i = 0; i < GLIDER_PERIOD; ++i) {
	const cell_t *p = up ? ptr : decr_ptr(ptr);
	cell_t origin = *p - glider->adjust[i][0];

	for (j = 0; j < GLIDER_LENGTH; ++j) {
	    if (*p != origin + glider->adjust[i][j]) { break; }
	    p = up ? incr_ptr(p) : decr_ptr(p);
	}

	if (j == GLIDER_LENGTH) {
	    coord_t y = GET_Y(origin + glider->adjust[i][0]);
	    if ((  up && y + GLIDER_MARGIN <= GET_Y(*p))
	     || (! up && y - GLIDER_MARGIN >= GET_Y(*p))) {
		*origin_ptr = origin;
		return i;
	    }
	    break;
	}
    }
    return -1;
}

static Boolean check_glider(
    int count, const cell_t *ptr1, const cell_t *ptr2,
    Boolean up, const struct glider *glider)
{
    int i, j;
    cell_t origin;
    const cell_t *p;

    i = is_glider(ptr1, up, glider, &origin);
    if (i >= 0) {
	for (j = 0; j < count; ++j) {
	    ++i;
	    if (i == GLIDER_PERIOD) {
		origin += glider->next;
		i = 0;
	    }
	}

	p = up ? ptr2 : decr_ptr(ptr2);
	for (j = 0; j < GLIDER_LENGTH; ++j) {
	    if (*p != origin + glider->adjust[i][j]) { break; }
	    p = up ? incr_ptr(p) : decr_ptr(p);
	}

	if (j == GLIDER_LENGTH) {
	    coord_t y = GET_Y(origin + glider->adjust[i][0]);
	    if ((  up && y + GLIDER_MARGIN <= GET_Y(*p))
	     || (! up && y - GLIDER_MARGIN >= GET_Y(*p))) {
		return true;
	    }
	}
    }

    return false;
}

static int check1(
    int count, const cell_t *begin, const cell_t *end,
    const cell_t *life_begin, const cell_t *life_end, offset_t life_length)
{
    int i, num_gliders = 0;
    const cell_t *p, *q;

    while (life_length > GLIDER_LENGTH) {
	if (check_glider(count, begin, life_begin, true, &glider_up_left)
	 || check_glider(count, begin, life_begin, true, &glider_up_right)) {
	    for (i = 0; i < GLIDER_LENGTH; ++i) {
		begin = incr_ptr(begin);
		life_begin = incr_ptr(life_begin);
	    }
	    life_length -= GLIDER_LENGTH;
	    ++num_gliders;
	    continue;
	}

	if (check_glider(count, end, life_end, false, &glider_down_left)
	 || check_glider(count, end, life_end, false, &glider_down_right)) {
	    for (i = 0; i < GLIDER_LENGTH; ++i) {
		end = decr_ptr(end);
		life_end = decr_ptr(life_end);
	    }
	    life_length -= GLIDER_LENGTH;
	    ++num_gliders;
	    continue;
	}

	break;
    }

    p = begin;
    q = life_begin;
    for ( ; p != end; p = incr_ptr(p), q = incr_ptr(q)) {
	if (*p != *q) { break; }
    }
    if (p == end && q == life_end) {
	return num_gliders;
    }
    return -1;
}

int check(const cell_t *life_begin, const cell_t *life_end, int *num_gliders)
{
    int count = 0;
    const cell_t *begin, *end;
    offset_t life_length = count_length(life_begin, life_end);

    begin = life_begin;
    end = life_end;
    while ((begin = (const cell_t *) *(end = decr_ptr(begin))) != NULL) {
	++count;
	if (count_length(begin, end) == life_length
	 && (*num_gliders = check1(
		count, begin, end, life_begin, life_end, life_length)) >= 0) {
	    return count;
	}
    }

    return 0;
}
