/* life.h */

/*
 * types and macros
 */

typedef long cell_t;
typedef long offset_t;
typedef short coord_t;
typedef short gen_t;

#define SENTINEL ((cell_t) 0x7fffffff)
#define BAD_CELL ((cell_t) 0x80008000)

#define CELL(x, y) ((((cell_t) (x) + (1 << 14)) <<  0) \
		  + (((cell_t) (y) + (1 << 14)) << 16))
#define GET_X(cell) ((coord_t) ((cell) >>  0) - (1 << 14))
#define GET_Y(cell) ((coord_t) ((cell) >> 16) - (1 << 14))

#define ADJUST(x, y) ((cell_t) (x) + ((cell_t) (y) << 16))


/*
 * function prototypes
 */

/* from life.c */

offset_t count_length(const cell_t *begin, const cell_t *end);

const cell_t *incr_ptr(const cell_t *ptr);
const cell_t *decr_ptr(const cell_t *ptr);

/* from load.c */

offset_t load(const char *str, long strlen, cell_t *tape);

/* from merge.c */

cell_t *merge(
    const cell_t *tape_begin, const cell_t *tape_end,
    const cell_t *life_begin, const cell_t *life_end,
    cell_t *out_begin, const cell_t *out_max);

/* from check.c */

int check(const cell_t *life_begin, const cell_t *life_end, int *num_gliders);
