/* merge.c */

#include <Pilot.h>

#include "life.h"

#define NUM_NEIGHBORS 9
#define CENTER_NO 4

static const cell_t adjust[NUM_NEIGHBORS] = {
    ADJUST(-1, -1), ADJUST( 0, -1), ADJUST( 1, -1),
    ADJUST(-1,  0), ADJUST( 0,  0), ADJUST( 1,  0),
    ADJUST(-1,  1), ADJUST( 0,  1), ADJUST( 1,  1)
};

static const Boolean rule[NUM_NEIGHBORS * 2] = {
    false, false, false, true,  false, false, false, false, false,
    false, false, true,  true,  false, false, false, false, false
};

cell_t *merge(
    const cell_t *tape_begin, const cell_t *tape_end,
    const cell_t *life_begin, const cell_t *life_end,
    cell_t *out_begin, const cell_t *out_max)
{
    const cell_t *ptrs[NUM_NEIGHBORS];

    struct head {
	cell_t cell;
	int tape_no;
	int dummy;
    } heads[NUM_NEIGHBORS + 1];

    cell_t current, *out = out_begin;
    int i, count;

    for (i = 0; i < NUM_NEIGHBORS; ++i) {
	ptrs[i] = life_begin;
	heads[i].cell = *life_begin + adjust[i];
	heads[i].tape_no = i;
    }
    heads[NUM_NEIGHBORS].cell = SENTINEL;

    current = heads[0].cell;
    count = 1;

    while (true) {
	int tape_no = heads[0].tape_no;
	const cell_t *ptr = ++ptrs[tape_no];
	cell_t cell;
	struct head *p;

	if (ptr == tape_end) {
	    ptr = ptrs[tape_no] = tape_begin;
	}

	if (ptr != life_end) {
	    cell = *ptr + adjust[tape_no];
	    for (p = heads; p[1].cell < cell; ++p) {
		*p = p[1];
	    }
	    p->cell = cell;
	    p->tape_no = tape_no;
	} else {
	    for (p = heads; p->cell != SENTINEL; ++p) {
		*p = p[1];
	    }
	    if (heads[0].cell == SENTINEL) {
		break;
	    }
	}

	if (heads[0].cell == current) {
	    if (heads[0].tape_no == CENTER_NO) {
		count += NUM_NEIGHBORS;
	    } else {
		++count;
	    }
	} else {
	    if (current & BAD_CELL) {
		return NULL;
	    }

	    if (rule[count]) {
		if (out == out_max) {
		    return NULL;
		}

		*out = current;
		if (++out == tape_end) {
		    out = (cell_t *) tape_begin;
		}
	    }

	    current = heads[0].cell;
	    if (heads[0].tape_no == CENTER_NO) {
		count = NUM_NEIGHBORS;
	    } else {
		count = 1;
	    }
	}
    }

    return out;
}
